import sys
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    Iterator,
    Mapping,
)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from .. import futures
from ..typing import S, T

__all__: list[str] = [
    "Pool",
    "ThreadPool",
    "AsyncResult",
    "ApplyResult",
    "MapResult",
]

class Pool:
    Executor: type[futures.MPIPoolExecutor]
    executor: futures.Executor
    def __init__(
        self,
        processes: int | None = None,
        initializer: Callable[..., object] | None = None,
        initargs: Iterable[Any] = (),
        **kwargs: Any,
    ) -> None: ...
    def apply(
        self,
        func: Callable[..., T],
        args: Iterable[Any] = (),
        kwds: Mapping[str, Any] = {},
    ) -> T: ...
    def apply_async(
        self,
        func: Callable[..., T],
        args: Iterable[Any] = (),
        kwds: Mapping[str, Any] = {},
        callback: Callable[[T], object] | None = None,
        error_callback: Callable[[BaseException], object] | None = None,
    ) -> AsyncResult[T]: ...
    def map(
        self,
        func: Callable[[S], T],
        iterable: Iterable[S],
        chunksize: int | None = None,
    ) -> list[T]: ...
    def map_async(
        self,
        func: Callable[[S], T],
        iterable: Iterable[S],
        chunksize: int | None = None,
        callback: Callable[[T], None] | None = None,
        error_callback: Callable[[BaseException], None] | None = None,
    ) -> MapResult[T]: ...
    def imap(
        self,
        func: Callable[[S], T],
        iterable: Iterable[S],
        chunksize: int = 1,
    ) -> Iterator[T]: ...
    def imap_unordered(
        self,
        func: Callable[[S], T],
        iterable: Iterable[S],
        chunksize: int = 1,
    ) -> Iterator[T]: ...
    def starmap(
        self,
        func: Callable[..., T],
        iterable: Iterable[Iterable[Any]],
        chunksize: int | None = None,
    ) -> list[T]: ...
    def starmap_async(
        self,
        func: Callable[..., T],
        iterable: Iterable[Iterable[Any]],
        chunksize: int | None = None,
        callback: Callable[[T], None] | None = None,
        error_callback: Callable[[BaseException], None] | None = None,
    ) -> MapResult[T]: ...
    def istarmap(
        self,
        func: Callable[..., T],
        iterable: Iterable[Iterable[Any]],
        chunksize: int = 1,
    ) -> Iterator[T]: ...
    def istarmap_unordered(
        self,
        func: Callable[..., T],
        iterable: Iterable[Iterable[Any]],
        chunksize: int = 1,
    ) -> Iterator[T]: ...
    def close(self) -> None: ...
    def terminate(self) -> None: ...
    def join(self) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, *args: object) -> None: ...

class ThreadPool(Pool):
    Executor: type[futures.ThreadPoolExecutor]

class AsyncResult(Generic[T]):
    future: futures.Future[T]
    def __init__(
        self,
        future: futures.Future[T],
        callback: Callable[[T], None] | None = None,
        error_callback: Callable[[BaseException], None] | None = None,
    ) -> None: ...
    def get(self, timeout: float | None = None) -> T: ...
    def wait(self, timeout: float | None = None) -> None: ...
    def ready(self) -> bool: ...
    def successful(self) -> bool: ...

class ApplyResult(AsyncResult[T]): ...
class MapResult(AsyncResult[list[T]]): ...



from typing import Sequence

__version__: str = ...
__author__: str = ...
__credits__: str = ...

from . import MPI

__all__: list[str] = ["MPI"]

class Rc:
    initialize: bool = True
    threads: bool = True
    thread_level: str = "multiple"
    finalize: bool | None = None
    fast_reduce: bool = True
    recv_mprobe: bool = True
    irecv_bufsz: int = 32768
    errors: str = "exception"
    def __init__(self, **kwargs: bool | str) -> None: ...
    def __setattr__(self, name: str, value: bool | str) -> None: ...
    def __call__(self, **kwargs: bool | str) -> None: ...

rc: Rc = ...

def get_include() -> str: ...
def get_config() -> dict[str, str]: ...
def profile(
    name: str,
    *,
    path: str | Sequence[str] | None = None,
) -> None: ...



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""Run Python code using ``mpi4py``.

Run Python code (scripts, modules, zip files) using the ``runpy``
module. In case of an unhandled exception, abort execution of the MPI
program by calling ``MPI.COMM_WORLD.Abort()``.
"""


def run_command_line(args=None):
    """Run command line ``[pyfile | -m mod | -c cmd | -] [arg] ...``.

    * ``pyfile`` : program read from script file
    * ``-m mod`` : run library module as a script
    * ``-c cmd`` : program passed in as a command string
    * ``-``      : program read from standard input (``sys.stdin``)
    * ``arg ...``: arguments passed to program in ``sys.argv[1:]``
    """
    # pylint: disable=import-outside-toplevel
    import os
    import sys
    from runpy import run_module, run_path

    def run_string(
        string,
        init_globals=None,
        run_name=None,
        filename="<string>",
        argv0="-c",
    ):
        from runpy import _run_module_code

        code = compile(string, filename, "exec", 0, True)
        kwargs = {"script_name": argv0}
        return _run_module_code(code, init_globals, run_name, **kwargs)

    sys.argv[:] = args if args is not None else sys.argv[1:]

    if sys.argv[0] == "-":
        cmd = sys.stdin.read()
        run_string(cmd, run_name="__main__", filename="<stdin>", argv0="-")
    elif sys.argv[0] == "-c":
        cmd = sys.argv.pop(1)  # Remove "cmd" from argument list
        run_string(cmd, run_name="__main__", filename="<string>", argv0="-c")
    elif sys.argv[0] == "-m":
        del sys.argv[0]  # Remove "-m" from argument list
        run_module(sys.argv[0], run_name="__main__", alter_sys=True)
    else:
        if not getattr(sys.flags, "isolated", 0):  # pragma: no branch
            dir_path = os.path.dirname(sys.argv[0])  # noqa: PTH120
            sys.path[0] = os.path.realpath(dir_path)  # Fix sys.path
        run_path(sys.argv[0], run_name="__main__")


def set_abort_status(status):
    """Terminate MPI execution environment at Python exit.

    Terminate MPI execution environment at Python exit by calling
    ``MPI.COMM_WORLD.Abort(status)``. This function should be called
    within an ``except`` block. Afterwards, exceptions should be
    re-raised.
    """
    # pylint: disable=import-outside-toplevel
    import sys

    if isinstance(status, SystemExit):
        status = status.code
    elif isinstance(status, KeyboardInterrupt):
        from _signal import SIGINT

        status = SIGINT + 128
    if not isinstance(status, int):
        status = 0 if status is None else 1
    pkg = __spec__.parent
    mpi = sys.modules.get(f"{pkg}.MPI")
    if mpi is not None and status:
        # pylint: disable=protected-access
        mpi._set_abort_status(status)


def main():
    """Entry-point for ``python -m mpi4py.run ...``."""
    # pylint: disable=too-many-statements
    # pylint: disable=import-outside-toplevel
    import os
    import sys

    def prefix():
        import pathlib

        prefix = pathlib.Path(__spec__.origin).parent
        print(prefix, file=sys.stdout)
        sys.exit(0)

    def version():
        from . import __version__

        package = __spec__.parent
        print(f"{package} {__version__}", file=sys.stdout)
        sys.exit(0)

    def mpi_library():
        from . import rc

        rc.initialize = rc.finalize = False
        from . import MPI

        def get_mpi_library_posix():
            import ctypes

            class DL_Info(ctypes.Structure):
                # pylint: disable=invalid-name
                # pylint: disable=too-few-public-methods
                # pylint: disable=missing-class-docstring
                _fields_ = [
                    ("dli_fname", ctypes.c_char_p),
                    ("dli_fbase", ctypes.c_void_p),
                    ("dli_sname", ctypes.c_char_p),
                    ("dli_saddr", ctypes.c_void_p),
                ]

            libc = ctypes.CDLL(None)
            dladdr = libc.dladdr
            dladdr.restype = ctypes.c_int
            dladdr.argtypes = [ctypes.c_void_p, ctypes.POINTER(DL_Info)]

            module = ctypes.CDLL(MPI.__file__)
            symbol = module.MPI_Init
            dlinfo = DL_Info()
            retval = dladdr(symbol, dlinfo)
            if not retval:  # pragma: no cover
                return None
            return os.fsdecode(dlinfo.dli_fname)

        def get_mpi_library_windows():  # pragma: no cover
            import ctypes
            import ctypes.wintypes

            # pylint: disable-next=invalid-name
            GetModuleHandleEx = ctypes.windll.kernel32.GetModuleHandleExW
            GetModuleHandleEx.restype = ctypes.wintypes.BOOL
            GetModuleHandleEx.argtypes = [
                ctypes.wintypes.DWORD,
                ctypes.wintypes.LPCVOID,
                ctypes.POINTER(ctypes.wintypes.HMODULE),
            ]
            # pylint: disable-next=invalid-name
            GetModuleFileName = ctypes.windll.kernel32.GetModuleFileNameW
            GetModuleFileName.restype = ctypes.wintypes.DWORD
            GetModuleFileName.argtypes = [
                ctypes.wintypes.HMODULE,
                ctypes.wintypes.LPWSTR,
                ctypes.wintypes.DWORD,
            ]

            vendor, _ = MPI.get_vendor()
            if vendor == "Intel MPI":
                libmpi = "impi.dll"
            elif vendor == "Microsoft MPI":
                libmpi = "msmpi.dll"
            else:
                return None
            module = ctypes.CDLL(libmpi)
            symbol = module.MPI_Init
            handle = ctypes.wintypes.HMODULE()
            retval = GetModuleHandleEx(0x2 | 0x4, symbol, handle)
            if not retval:
                return None
            wsbuf = ctypes.create_unicode_buffer(1024)
            retval = GetModuleFileName(handle, wsbuf, ctypes.sizeof(wsbuf))
            if not retval:
                return None
            return wsbuf.value

        if os.name == "posix":
            libmpi = get_mpi_library_posix()
            print(libmpi, file=sys.stdout)
            sys.exit(0)
        elif os.name == "nt":  # pragma: no cover
            libmpi = get_mpi_library_windows()
            print(libmpi, file=sys.stdout)
            sys.exit(0)
        else:  # pragma: no cover
            print(f"Unsupported platform: {sys.platform}", file=sys.stderr)
            sys.exit(1)

    def usage(errmess=None):
        dedent = __import__("textwrap").dedent
        python = __import__("pathlib").Path(sys.executable).name
        program = __spec__.name

        cmdline = dedent(f"""
        usage: {python} -m {program} [options] <pyfile> [arg] ...
           or: {python} -m {program} [options] -m <mod> [arg] ...
           or: {python} -m {program} [options] -c <cmd> [arg] ...
           or: {python} -m {program} [options] - [arg] ...
        """).strip()

        helptip = dedent(f"""
        Try `{python} -m {program} -h` for more information.
        """).strip()

        options = dedent("""
        options:
          --prefix             show install path and exit
          --version            show version number and exit
          --mpi-library        show MPI library path and exit
          --mpi-std-version    show MPI standard version and exit
          --mpi-lib-version    show MPI library version and exit
          -h|--help            show this help message and exit
          -rc <key=value,...>  set 'mpi4py.rc.key=value'
        """).strip()

        if errmess:
            print(errmess, file=sys.stderr)
            print(cmdline, file=sys.stderr)
            print(helptip, file=sys.stderr)
            sys.exit(1)
        else:
            print(cmdline, file=sys.stdout)
            print(options, file=sys.stdout)
            sys.exit(0)

    def mpi_std_version():
        from . import rc

        rc.initialize = rc.finalize = False
        from . import MPI

        version = ".".join(map(str, (MPI.VERSION, MPI.SUBVERSION)))
        rtversion = ".".join(map(str, MPI.Get_version()))
        note = f" (runtime: MPI {rtversion})" if rtversion != version else ""
        print(f"MPI {version}{note}", file=sys.stdout)
        sys.exit(0)

    def mpi_lib_version():
        from . import rc

        rc.initialize = rc.finalize = False
        from . import MPI

        library_version = MPI.Get_library_version()
        print(library_version, file=sys.stdout)
        sys.exit(0)

    def parse_command_line(args=None):
        # pylint: disable=too-many-branches

        class Options:
            # pylint: disable=too-few-public-methods
            # pylint: disable=missing-class-docstring
            rc_args = {}

        def poparg(args):
            if len(args) < 2 or args[1].startswith("-"):
                usage("Argument expected for option: " + args[0])
            return args.pop(1)

        options = Options()
        args = sys.argv[1:] if args is None else args[:]
        while args and args[0].startswith("-"):
            arg0 = args[0]
            if arg0 in ("-m", "-c", "-"):
                break  # Stop processing options
            if arg0 in ("-h", "-help", "--help"):
                usage()  # Print help and exit
            if arg0 in ("-prefix", "--prefix"):
                prefix()  # Print install path and exit
            if arg0 in ("-version", "--version"):
                version()  # Print version number and exit
            if arg0 in ("-mpi-library", "--mpi-library"):
                mpi_library()  # Print MPI library path and exit
            if arg0 in ("-mpi-std-version", "--mpi-std-version"):
                mpi_std_version()  # Print MPI standard version and exit
            if arg0 in ("-mpi-lib-version", "--mpi-lib-version"):
                mpi_lib_version()  # Print MPI library version and exit
            if arg0.startswith("--"):
                if "=" in arg0:
                    opt, _, arg = arg0[1:].partition("=")
                    if opt == "-rc":
                        arg0, args[1:1] = opt, [arg]
                else:
                    arg0 = arg0[1:]
            if arg0 == "-rc":
                from ast import literal_eval

                for entry in poparg(args).split(","):
                    key, _, val = entry.partition("=")
                    if not key or not val:
                        usage("Cannot parse rc option: " + entry)
                    try:
                        val = literal_eval(val)
                    except ValueError:
                        pass
                    options.rc_args[key] = val
            else:
                usage("Unknown option: " + args[0])
            del args[0]
        # Check remaining args and return to caller
        if not args:
            usage("No path specified for execution")
        if args[0] in ("-m", "-c") and len(args) < 2:
            usage("Argument expected for option: " + args[0])
        return options, args

    def bootstrap(options):
        if options.rc_args:  # Set mpi4py.rc parameters
            from . import rc

            rc(**options.rc_args)

    # Parse and process command line options
    options, args = parse_command_line()
    bootstrap(options)

    # Run user code. In case of an unhandled exception, abort
    # execution of the MPI program by calling 'MPI_Abort()'.
    try:
        run_command_line(args)
    except SystemExit as exc:
        set_abort_status(exc)
        raise
    except KeyboardInterrupt as exc:
        set_abort_status(exc)
        raise
    except BaseException:
        set_abort_status(1)
        raise


if __name__ == "__main__":
    main()



from typing import Callable, Generic, overload

from ..typing import S, T, U
from ._base import Future

class ThenableFuture(Future[T], Generic[T]):
    def then(
        self,
        on_success: Callable[[T], T | Future[T]] | None = None,
        on_failure: Callable[[BaseException], T | BaseException] | None = None,
    ) -> ThenableFuture[T]: ...
    def catch(
        self,
        on_failure: Callable[[BaseException], T | BaseException] | None = None,
    ) -> ThenableFuture[T]: ...

@overload
def then(
    future: Future[T],
    on_success: None = None,
    on_failure: None = None,
) -> Future[T]: ...
@overload
def then(
    future: Future[T],
    on_success: Callable[[T], S | Future[S]],
    on_failure: None = None,
) -> Future[S]: ...
@overload
def then(
    future: Future[T],
    on_success: None = None,
    *,
    on_failure: Callable[[BaseException], BaseException | U | Future[U]],
) -> Future[T | U]: ...
@overload
def then(
    future: Future[T],
    on_success: Callable[[T], S | Future[S]],
    on_failure: Callable[[BaseException], BaseException | U | Future[U]],
) -> Future[S | U]: ...
@overload
def catch(
    future: Future[T],
    on_failure: None = None,
) -> Future[T]: ...
@overload
def catch(
    future: Future[T],
    on_failure: Callable[[BaseException], BaseException | U | Future[U]],
) -> Future[T | U]: ...



from typing import Callable, Collection, overload

from ..typing import T, U, V
from ._base import Future

__all__: list[str] = [
    "collect",
    "compose",
]

def collect(
    fs: Collection[Future[T]],
) -> Future[list[T]]: ...
@overload
def compose(
    future: Future[T],
    resulthook: None = None,
    excepthook: None = None,
) -> Future[T]: ...
@overload
def compose(
    future: Future[T],
    resulthook: Callable[[T], U],
    excepthook: None = None,
) -> Future[U]: ...
@overload
def compose(
    future: Future[T],
    resulthook: None = None,
    *,
    excepthook: Callable[[BaseException], BaseException | V],
) -> Future[T | V]: ...
@overload
def compose(
    future: Future[T],
    resulthook: Callable[[T], U],
    excepthook: Callable[[BaseException], BaseException | V],
) -> Future[U | V]: ...



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""Synchronization utilities."""

import array as _array
import time as _time

from .. import MPI

__all__ = [
    "Sequential",
    "Counter",
    "Mutex",
    "Condition",
    "Semaphore",
]


class Sequential:
    """Sequential execution."""

    def __init__(self, comm, tag=0):
        """Initialize sequential execution.

        Args:
            comm: Intracommunicator context.
            tag: Tag for point-to-point communication.

        """
        self.comm = comm
        self.tag = int(tag)

    def __enter__(self):
        """Enter sequential execution."""
        self.begin()
        return self

    def __exit__(self, *exc):
        """Exit sequential execution."""
        self.end()

    def begin(self):
        """Begin sequential execution."""
        comm = self.comm
        size = comm.Get_size()
        if size == 1:
            return
        rank = comm.Get_rank()
        buf = (bytearray(), 0, MPI.BYTE)
        tag = self.tag
        if rank != 0:
            comm.Recv(buf, rank - 1, tag)

    def end(self):
        """End sequential execution."""
        comm = self.comm
        size = comm.Get_size()
        if size == 1:
            return
        rank = comm.Get_rank()
        buf = (bytearray(), 0, MPI.BYTE)
        tag = self.tag
        if rank != size - 1:
            comm.Send(buf, rank + 1, tag)


class Counter:
    """Global counter."""

    def __init__(
        self,
        start=0,
        step=1,
        *,
        typecode="i",
        comm=MPI.COMM_SELF,
        info=MPI.INFO_NULL,
        root=0,
    ):
        """Initialize global counter.

        Args:
            start: Start value.
            step: Increment value.
            typecode: Type code as defined in the `array` module.
            comm: Intracommunicator context.
            info: Info object for RMA context creation.
            root: Process rank holding the counter memory.

        """
        # pylint: disable=too-many-arguments
        datatype = MPI.Datatype.fromcode(typecode)
        typechar = datatype.typechar
        rank = comm.Get_rank()
        count = 1 if rank == root else 0
        unitsize = datatype.Get_size()
        window = MPI.Win.Allocate(count * unitsize, unitsize, info, comm)
        self._start = start
        self._step = step
        self._window = window
        self._typechar = typechar
        self._location = (root, 0)
        self._comm = comm

        init = _array.array(typechar, [start] * count)
        window.Lock(rank, MPI.LOCK_SHARED)
        window.Accumulate(init, rank, op=MPI.REPLACE)
        window.Unlock(rank)
        comm.Barrier()

    def __iter__(self):
        """Implement ``iter(self)``."""
        return self

    def __next__(self):
        """Implement ``next(self)``."""
        return self.next()

    def next(self, incr=None):
        """Return current value and increment.

        Args:
            incr: Increment value.

        Returns:
            The counter value before incrementing.

        """
        if not self._window:
            raise RuntimeError("counter already freed")

        window = self._window
        typechar = self._typechar
        root, disp = self._location

        incr = incr if incr is not None else self._step
        incr = _array.array(typechar, [incr])
        prev = _array.array(typechar, [0])
        op = MPI.SUM if incr[0] != 0 else MPI.NO_OP
        window.Lock(root, MPI.LOCK_SHARED)
        window.Fetch_and_op(incr, prev, root, disp, op)
        window.Unlock(root)
        return prev[0]

    def free(self):
        """Free counter resources."""
        window = self._window
        self._window = MPI.WIN_NULL
        self._comm = MPI.COMM_NULL
        window.free()


class Mutex:
    """Mutual exclusion."""

    def __init__(
        self,
        *,
        recursive=False,
        comm=MPI.COMM_SELF,
        info=MPI.INFO_NULL,
    ):
        """Initialize mutex object.

        Args:
            comm: Intracommunicator context.
            recursive: Whether to allow recursive acquisition.
            info: Info object for RMA context creation.

        """
        null_rank, tail_rank = MPI.PROC_NULL, 0

        rank = comm.Get_rank()
        count = 3 if rank == tail_rank else 2
        unitsize = MPI.INT.Get_size()
        window = MPI.Win.Allocate(count * unitsize, unitsize, info, comm)
        self._recursive = bool(recursive)
        self._window = window
        self._comm = comm

        init = [False, null_rank, null_rank][:count]
        init = _array.array("i", init)
        window.Lock(rank, MPI.LOCK_SHARED)
        window.Accumulate(init, rank, op=MPI.REPLACE)
        window.Unlock(rank)
        comm.Barrier()

    def _acquire(self, blocking=True):
        null_rank, tail_rank = MPI.PROC_NULL, 0
        lock_id, next_id, tail_id = (0, 1, 2)

        window = self._window
        self_rank = window.group_rank
        window.Lock_all()

        rank = _array.array("i", [self_rank])
        null = _array.array("i", [null_rank])
        prev = _array.array("i", [null_rank])
        window.Accumulate(null, self_rank, next_id, MPI.REPLACE)
        if blocking:
            window.Fetch_and_op(rank, prev, tail_rank, tail_id, MPI.REPLACE)
        else:
            window.Compare_and_swap(rank, null, prev, tail_rank, tail_id)
        window.Flush(tail_rank)
        locked = int(prev[0] == null_rank)
        if blocking and not locked:
            # Add ourselves to the waiting queue
            window.Accumulate(rank, prev[0], next_id, MPI.REPLACE)
            # Spin until we are given the lock
            locked = self._spinloop(lock_id, 0)

        # Set the local lock flag
        flag = _array.array("i", [locked])
        window.Accumulate(flag, self_rank, lock_id, MPI.REPLACE)

        window.Unlock_all()
        return bool(locked)

    def _release(self):
        null_rank, tail_rank = MPI.PROC_NULL, 0
        lock_id, next_id, tail_id = (0, 1, 2)

        window = self._window
        self_rank = window.group_rank
        window.Lock_all()

        rank = _array.array("i", [self_rank])
        null = _array.array("i", [null_rank])
        prev = _array.array("i", [null_rank])
        window.Compare_and_swap(null, rank, prev, tail_rank, tail_id)
        window.Flush(tail_rank)
        if prev[0] != rank[0]:
            # Spin until the next process notify us
            next_rank = self._spinloop(next_id, null_rank)
            # Pass the lock over to the next process
            true = _array.array("i", [True])
            window.Accumulate(true, next_rank, lock_id, MPI.REPLACE)

        # Set the local lock flag
        false = _array.array("i", [False])
        window.Accumulate(false, self_rank, lock_id, MPI.REPLACE)

        window.Unlock_all()

    def _count_fetch_and_op(self, value, op):
        lock_id = 0
        window = self._window
        self_rank = window.group_rank
        incr = _array.array("i", [value])
        prev = _array.array("i", [0])
        window.Lock(self_rank, MPI.LOCK_SHARED)
        window.Fetch_and_op(incr, prev, self_rank, lock_id, op)
        window.Unlock(self_rank)
        return prev[0]

    def _acquire_restore(self, state):
        self._acquire()
        if self._recursive:
            self._count_fetch_and_op(state, MPI.REPLACE)

    def _release_save(self):
        state = None
        if self._recursive:
            state = self._count_fetch_and_op(0, MPI.NO_OP)
        self._release()
        return state

    def _spinloop(self, index, sentinel):
        window = self._window
        return _rma_spinloop(window, "i", index, sentinel)

    def __enter__(self):
        """Acquire mutex."""
        self.acquire()
        return self

    def __exit__(self, *exc):
        """Release mutex."""
        self.release()

    def acquire(self, blocking=True):
        """Acquire mutex, blocking or non-blocking.

        Args:
            blocking: If `True`, block until the mutex is held.

        Returns:
            `True` if the mutex is held, `False` otherwise.

        """
        if not self._window:
            raise RuntimeError("mutex already freed")
        if self.locked():
            if self._recursive:
                self._count_fetch_and_op(+1, MPI.SUM)
                return True
            raise RuntimeError("cannot acquire already held mutex")
        return self._acquire(blocking)

    def release(self):
        """Release mutex."""
        if not self._window:
            raise RuntimeError("mutex already freed")
        if not self.locked():
            raise RuntimeError("cannot release unheld mutex")
        if self._recursive:
            if self._count_fetch_and_op(-1, MPI.SUM) > 1:
                return
        self._release()

    def locked(self):
        """Return whether the mutex is held."""
        if not self._window:
            raise RuntimeError("mutex already freed")
        lock_id = 0
        memory = memoryview(self._window).cast("i")
        return bool(memory[lock_id])

    def count(self):
        """Return the recursion count."""
        if not self._window:
            raise RuntimeError("mutex already freed")
        return self._count_fetch_and_op(0, MPI.NO_OP)

    def free(self):
        """Free mutex resources."""
        if self._window:
            if self.locked():
                self._release()
        window = self._window
        self._window = MPI.WIN_NULL
        self._comm = MPI.COMM_NULL
        window.free()


class Condition:
    """Condition variable."""

    def __init__(
        self,
        mutex=None,
        *,
        recursive=True,
        comm=MPI.COMM_SELF,
        info=MPI.INFO_NULL,
    ):
        """Initialize condition variable.

        Args:
            mutex: Mutual exclusion object.
            recursive: Whether to allow recursive acquisition.
            comm: Intracommunicator context.
            info: Info object for RMA context creation.

        """
        if mutex is None:
            self._mutex = Mutex(recursive=recursive, comm=comm, info=info)
            self._mutex_free = self._mutex.free
        else:
            self._mutex = mutex
            self._mutex_free = lambda: None
            comm = mutex._comm  # pylint disable=protected-access

        null_rank, tail_rank = MPI.PROC_NULL, 0

        rank = comm.Get_rank()
        count = 3 if rank == tail_rank else 2
        unitsize = MPI.INT.Get_size()
        window = MPI.Win.Allocate(count * unitsize, unitsize, info, comm)
        self._window = window
        self._comm = comm

        init = [0, null_rank, null_rank][:count]
        init = _array.array("i", init)
        window.Lock(rank, MPI.LOCK_SHARED)
        window.Accumulate(init, rank, op=MPI.REPLACE)
        window.Unlock(rank)
        comm.Barrier()

    def _enqueue(self, process):
        # pylint: disable=redefined-builtin
        null_rank, tail_rank = MPI.PROC_NULL, 0
        next_id, tail_id = (1, 2)
        window = self._window

        rank = _array.array("i", [process])
        prev = _array.array("i", [null_rank])
        next = _array.array("i", [process])  # noqa: A001

        window.Lock_all()
        window.Fetch_and_op(rank, prev, tail_rank, tail_id, MPI.REPLACE)
        window.Flush(tail_rank)
        if prev[0] != null_rank:
            window.Fetch_and_op(rank, next, prev[0], next_id, MPI.REPLACE)
            window.Flush(prev[0])
        window.Accumulate(next, rank[0], next_id, MPI.REPLACE)
        window.Unlock_all()

    def _dequeue(self, maxnumprocs):
        # pylint: disable=redefined-builtin
        null_rank, tail_rank = MPI.PROC_NULL, 0
        next_id, tail_id = (1, 2)
        window = self._window

        null = _array.array("i", [null_rank])
        prev = _array.array("i", [null_rank])
        next = _array.array("i", [null_rank])  # noqa: A001

        processes = []
        maxnumprocs = max(0, min(maxnumprocs, window.group_size))
        window.Lock_all()
        window.Fetch_and_op(null, prev, tail_rank, tail_id, MPI.NO_OP)
        window.Flush(tail_rank)
        if prev[0] != null_rank:
            empty = False
            window.Fetch_and_op(null, next, prev[0], next_id, MPI.NO_OP)
            window.Flush(prev[0])
            while len(processes) < maxnumprocs and not empty:
                rank = next[0]
                processes.append(rank)
                window.Fetch_and_op(null, next, rank, next_id, MPI.NO_OP)
                window.Flush(rank)
                empty = processes[0] == next[0]
            if not empty:
                window.Accumulate(next, prev[0], next_id, MPI.REPLACE)
            else:
                window.Accumulate(null, tail_rank, tail_id, MPI.REPLACE)
        window.Unlock_all()
        return processes

    def _sleep(self):
        flag_id = 0
        window = self._window
        window.Lock_all()
        _rma_spinloop(window, "i", flag_id, 0, reset=True)
        window.Unlock_all()

    def _wakeup(self, processes):
        flag_id = 0
        window = self._window
        flag = _array.array("i", [1])
        window.Lock_all()
        for rank in processes:
            window.Accumulate(flag, rank, flag_id, MPI.REPLACE)
        window.Unlock_all()

    def _release_save(self):
        # pylint: disable=protected-access
        return self._mutex._release_save()

    def _acquire_restore(self, state):
        # pylint: disable=protected-access
        self._mutex._acquire_restore(state)

    def _mutex_reset(self):
        # pylint: disable=protected-access
        if self._mutex._window:
            if self._mutex.locked():
                self._mutex._release()

    def __enter__(self):
        """Acquire the underlying mutex."""
        self.acquire()
        return self

    def __exit__(self, *exc):
        """Release the underlying mutex."""
        self.release()

    def acquire(self, blocking=True):
        """Acquire the underlying mutex."""
        if not self._window:
            raise RuntimeError("condition already freed")
        return self._mutex.acquire(blocking)

    def release(self):
        """Release the underlying mutex."""
        if not self._window:
            raise RuntimeError("condition already freed")
        self._mutex.release()

    def locked(self):
        """Return whether the underlying mutex is held."""
        return self._mutex.locked()

    def wait(self):
        """Wait until notified by another process.

        Returns:
            Always `True`.

        """
        if not self._window:
            raise RuntimeError("condition already freed")
        if not self.locked():
            raise RuntimeError("cannot wait on unheld mutex")
        self._enqueue(self._window.group_rank)
        state = self._release_save()
        self._sleep()
        self._acquire_restore(state)
        return True

    def wait_for(self, predicate):
        """Wait until a predicate evaluates to `True`.

        Args:
            predicate: callable returning a boolean.

        Returns:
            The result of predicate once it evaluates to `True`.

        """
        result = predicate()
        while not result:
            self.wait()
            result = predicate()
        return result

    def notify(self, n=1):
        """Wake up one or more processes waiting on this condition.

        Args:
            n: Maximum number of processes to wake up.

        Returns:
            The actual number of processes woken up.

        """
        if not self._window:
            raise RuntimeError("condition already freed")
        if not self.locked():
            raise RuntimeError("cannot notify on unheld mutex")
        processes = self._dequeue(n)
        numprocs = len(processes)
        self._wakeup(processes)
        return numprocs

    def notify_all(self):
        """Wake up all processes waiting on this condition.

        Returns:
            The actual number of processes woken up.

        """
        return self.notify((1 << 31) - 1)

    def free(self):
        """Free condition resources."""
        self._mutex_reset()
        self._mutex_free()
        window = self._window
        self._window = MPI.WIN_NULL
        self._comm = MPI.COMM_NULL
        window.free()


class Semaphore:
    """Semaphore object."""

    def __init__(
        self,
        value=1,
        *,
        bounded=True,
        comm=MPI.COMM_SELF,
        info=MPI.INFO_NULL,
    ):
        """Initialize semaphore object.

        Args:
            value: Initial value for internal counter.
            bounded: Bound internal counter to initial value.
            comm: Intracommunicator context.
            info: Info object for RMA context creation.

        """
        if value < 0:
            raise ValueError("initial value must be non-negative")
        self._bounded = bool(bounded)
        self._counter = Counter(value, comm=comm, info=info)
        self._condvar = Condition(recursive=False, comm=comm, info=info)
        self._comm = comm

    def __enter__(self):
        """Acquire semaphore."""
        self.acquire()
        return self

    def __exit__(self, *exc):
        """Release semaphore."""
        self.release()

    def acquire(self, blocking=True):
        """Acquire semaphore, decrementing the internal counter by one.

        Args:
            blocking: If `True`, block until the semaphore is acquired.

        Returns:
            `True` if the semaphore is acquired, `False` otherwise.

        """
        with self._condvar:
            while self._counter.next(0) == 0:
                if not blocking:
                    return False
                self._condvar.wait()
            self._counter.next(-1)
            return True

    def release(self, n=1):
        """Release semaphore, incrementing the internal counter by one or more.

        Args:
            n: Increment for the internal counter.

        """
        if n < 1:
            raise ValueError("increment must be one or more")
        with self._condvar:
            if self._bounded:
                # pylint: disable=protected-access
                current = self._counter.next(0)
                initial = self._counter._start
                if current + n > initial:
                    raise ValueError("semaphore released too many times")
            self._counter.next(n)
            self._condvar.notify(n)

    def free(self):
        """Free semaphore resources."""
        self._counter.free()
        self._condvar.free()
        self._comm = MPI.COMM_NULL


_BACKOFF_DELAY_MAX = 1 / 1024
_BACKOFF_DELAY_MIN = _BACKOFF_DELAY_MAX / 1024
_BACKOFF_DELAY_INIT = 0.0
_BACKOFF_DELAY_RATIO = 2.0


def _new_backoff(
    delay_max=_BACKOFF_DELAY_MAX,
    delay_min=_BACKOFF_DELAY_MIN,
    delay_init=_BACKOFF_DELAY_INIT,
    delay_ratio=_BACKOFF_DELAY_RATIO,
):
    def backoff_iterator():
        delay = delay_init
        while True:
            _time.sleep(delay)
            delay = min(delay_max, max(delay_min, delay * delay_ratio))
            yield

    backoff = backoff_iterator()
    return lambda: next(backoff)


def _rma_progress(window):
    window.Flush(window.group_rank)


def _rma_spinloop(
    window,
    typecode,
    index,
    sentinel,
    reset=False,
    backoff=None,
    progress=None,
):  # pylint: disable=too-many-arguments,too-many-positional-arguments
    memory = memoryview(window).cast(typecode)
    backoff = backoff or _new_backoff()
    progress = progress or _rma_progress
    window.Sync()
    while memory[index] == sentinel:
        backoff()
        progress(window)
        window.Sync()
    value = memory[index]
    if reset:
        memory[index] = sentinel
    return value



from typing import Any

from numpy import dtype
from numpy.typing import DTypeLike

from ..MPI import Datatype

# mypy: disable-error-code="no-any-unimported"
def from_numpy_dtype(dtype: DTypeLike) -> Datatype: ...
def to_numpy_dtype(datatype: Datatype) -> dtype[Any]: ...



from typing import Sequence

from .MPI import Intracomm

def helloworld(
    comm: Intracomm, args: Sequence[str] | None = None, verbose: bool = True
) -> str: ...
def ringtest(
    comm: Intracomm, args: Sequence[str] | None = None, verbose: bool = True
) -> float: ...
def pingpong(
    comm: Intracomm, args: Sequence[str] | None = None, verbose: bool = True
) -> list[tuple[int, float, float]]: ...
def futures(
    comm: Intracomm, args: Sequence[str] | None = None, verbose: bool = True
) -> list[tuple[int, float, float]]: ...
def main(args: Sequence[str] | None = ...) -> None: ...



from ._base import (
    ALL_COMPLETED as ALL_COMPLETED,
    FIRST_COMPLETED as FIRST_COMPLETED,
    FIRST_EXCEPTION as FIRST_EXCEPTION,
    BrokenExecutor as BrokenExecutor,
    CancelledError as CancelledError,
    Executor as Executor,
    Future as Future,
    InvalidStateError as InvalidStateError,
    TimeoutError as TimeoutError,  # noqa: A004
    as_completed as as_completed,
    wait as wait,
)
from .pool import (
    MPICommExecutor as MPICommExecutor,
    MPIPoolExecutor as MPIPoolExecutor,
    ProcessPoolExecutor as ProcessPoolExecutor,
    ThreadPoolExecutor as ThreadPoolExecutor,
    get_comm_workers as get_comm_workers,
)
from .util import (
    collect as collect,
    compose as compose,
)

__all__: list[str] = [
    "Future",
    "Executor",
    "wait",
    "FIRST_COMPLETED",
    "FIRST_EXCEPTION",
    "ALL_COMPLETED",
    "as_completed",
    "CancelledError",
    "TimeoutError",
    "InvalidStateError",
    "BrokenExecutor",
    "MPIPoolExecutor",
    "MPICommExecutor",
    "ThreadPoolExecutor",
    "ProcessPoolExecutor",
    "get_comm_workers",
    "collect",
    "compose",
]
